"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs = __importStar(require("fs"));
const tmp = __importStar(require("tmp-promise"));
const stream = __importStar(require("stream"));
const utils_1 = require("./utils");
const config_variables_1 = require("./config-variables");
const util_1 = require("util");
const url_1 = require("url");
const perf_hooks_1 = require("perf_hooks");
const upload_status_reporter_1 = require("./upload-status-reporter");
const core_1 = require("@actions/core");
const http_manager_1 = require("./http-manager");
const upload_gzip_1 = require("./upload-gzip");
const stat = util_1.promisify(fs.stat);
class UploadHttpClient {
    constructor() {
        this.uploadHttpManager = new http_manager_1.HttpManager(config_variables_1.getUploadFileConcurrency());
        this.statusReporter = new upload_status_reporter_1.UploadStatusReporter();
    }
    /**
     * Creates a file container for the new artifact in the remote blob storage/file service
     * @param {string} artifactName Name of the artifact being created
     * @returns The response from the Artifact Service if the file container was successfully created
     */
    createArtifactInFileContainer(artifactName) {
        return __awaiter(this, void 0, void 0, function* () {
            const parameters = {
                Type: 'actions_storage',
                Name: artifactName
            };
            const data = JSON.stringify(parameters, null, 2);
            const artifactUrl = utils_1.getArtifactUrl();
            // use the first client from the httpManager, `keep-alive` is not used so the connection will close immediatly
            const client = this.uploadHttpManager.getClient(0);
            const requestOptions = utils_1.getRequestOptions('application/json', false, false);
            const rawResponse = yield client.post(artifactUrl, data, requestOptions);
            const body = yield rawResponse.readBody();
            if (utils_1.isSuccessStatusCode(rawResponse.message.statusCode) && body) {
                return JSON.parse(body);
            }
            else {
                // eslint-disable-next-line no-console
                console.log(rawResponse);
                throw new Error(`Unable to create a container for the artifact ${artifactName}`);
            }
        });
    }
    /**
     * Concurrently upload all of the files in chunks
     * @param {string} uploadUrl Base Url for the artifact that was created
     * @param {SearchResult[]} filesToUpload A list of information about the files being uploaded
     * @returns The size of all the files uploaded in bytes
     */
    uploadArtifactToFileContainer(uploadUrl, filesToUpload, options) {
        return __awaiter(this, void 0, void 0, function* () {
            const FILE_CONCURRENCY = config_variables_1.getUploadFileConcurrency();
            const MAX_CHUNK_SIZE = config_variables_1.getUploadChunkSize();
            core_1.debug(`File Concurrency: ${FILE_CONCURRENCY}, and Chunk Size: ${MAX_CHUNK_SIZE}`);
            const parameters = [];
            // by default, file uploads will continue if there is an error unless specified differently in the options
            let continueOnError = true;
            if (options) {
                if (options.continueOnError === false) {
                    continueOnError = false;
                }
            }
            // prepare the necessary parameters to upload all the files
            for (const file of filesToUpload) {
                const resourceUrl = new url_1.URL(uploadUrl);
                resourceUrl.searchParams.append('itemPath', file.uploadFilePath);
                parameters.push({
                    file: file.absoluteFilePath,
                    resourceUrl: resourceUrl.toString(),
                    maxChunkSize: MAX_CHUNK_SIZE,
                    continueOnError
                });
            }
            const parallelUploads = [...new Array(FILE_CONCURRENCY).keys()];
            const failedItemsToReport = [];
            let currentFile = 0;
            let completedFiles = 0;
            let uploadFileSize = 0;
            let totalFileSize = 0;
            let abortPendingFileUploads = false;
            this.statusReporter.setTotalNumberOfFilesToUpload(filesToUpload.length);
            this.statusReporter.start();
            // only allow a certain amount of files to be uploaded at once, this is done to reduce potential errors
            yield Promise.all(parallelUploads.map((index) => __awaiter(this, void 0, void 0, function* () {
                while (currentFile < filesToUpload.length) {
                    const currentFileParameters = parameters[currentFile];
                    currentFile += 1;
                    if (abortPendingFileUploads) {
                        failedItemsToReport.push(currentFileParameters.file);
                        continue;
                    }
                    const startTime = perf_hooks_1.performance.now();
                    const uploadFileResult = yield this.uploadFileAsync(index, currentFileParameters);
                    core_1.debug(`File: ${++completedFiles}/${filesToUpload.length}. ${currentFileParameters.file} took ${(perf_hooks_1.performance.now() - startTime).toFixed(3)} milliseconds to finish upload`);
                    uploadFileSize += uploadFileResult.successfullUploadSize;
                    totalFileSize += uploadFileResult.totalSize;
                    if (uploadFileResult.isSuccess === false) {
                        failedItemsToReport.push(currentFileParameters.file);
                        if (!continueOnError) {
                            // existing uploads will be able to finish however all pending uploads will fail fast
                            abortPendingFileUploads = true;
                        }
                    }
                    this.statusReporter.incrementProcessedCount();
                }
            })));
            this.statusReporter.stop();
            // done uploading, safety dispose all connections
            this.uploadHttpManager.disposeAndReplaceAllClients();
            core_1.info(`Total size of all the files uploaded is ${uploadFileSize} bytes`);
            return {
                uploadSize: uploadFileSize,
                totalSize: totalFileSize,
                failedItems: failedItemsToReport
            };
        });
    }
    /**
     * Asynchronously uploads a file. The file is compressed and uploaded using GZip if it is determined to save space.
     * If the upload file is bigger than the max chunk size it will be uploaded via multiple calls
     * @param {number} httpClientIndex The index of the httpClient that is being used to make all of the calls
     * @param {UploadFileParameters} parameters Information about the file that needs to be uploaded
     * @returns The size of the file that was uploaded in bytes along with any failed uploads
     */
    uploadFileAsync(httpClientIndex, parameters) {
        return __awaiter(this, void 0, void 0, function* () {
            const totalFileSize = (yield stat(parameters.file)).size;
            let offset = 0;
            let isUploadSuccessful = true;
            let failedChunkSizes = 0;
            let uploadFileSize = 0;
            let isGzip = true;
            // the file that is being uploaded is less than 64k in size, to increase thoroughput and to minimize disk I/O
            // for creating a new GZip file, an in-memory buffer is used for compression
            if (totalFileSize < 65536) {
                const buffer = yield upload_gzip_1.createGZipFileInBuffer(parameters.file);
                let uploadStream;
                if (totalFileSize < buffer.byteLength) {
                    // compression did not help with reducing the size, use a readable stream from the original file for upload
                    uploadStream = fs.createReadStream(parameters.file);
                    isGzip = false;
                    uploadFileSize = totalFileSize;
                }
                else {
                    // create a readable stream using a PassThrough stream that is both readable and writable
                    const passThrough = new stream.PassThrough();
                    passThrough.end(buffer);
                    uploadStream = passThrough;
                    uploadFileSize = buffer.byteLength;
                }
                const result = yield this.uploadChunk(httpClientIndex, parameters.resourceUrl, uploadStream, 0, uploadFileSize - 1, uploadFileSize, isGzip, totalFileSize);
                if (!result) {
                    // chunk failed to upload
                    isUploadSuccessful = false;
                    failedChunkSizes += uploadFileSize;
                    core_1.warning(`Aborting upload for ${parameters.file} due to failure`);
                }
                return {
                    isSuccess: isUploadSuccessful,
                    successfullUploadSize: uploadFileSize - failedChunkSizes,
                    totalSize: totalFileSize
                };
            }
            else {
                // the file that is being uploaded is greater than 64k in size, a temprorary file gets created on disk using the
                // npm tmp-promise package and this file gets used during compression for the GZip file that gets created
                return tmp
                    .file()
                    .then((tmpFile) => __awaiter(this, void 0, void 0, function* () {
                    // create a GZip file of the original file being uploaded, the original file should not be modified in any way
                    uploadFileSize = yield upload_gzip_1.createGZipFileOnDisk(parameters.file, tmpFile.path);
                    let uploadFilePath = tmpFile.path;
                    // compression did not help with size reduction, use the original file for upload and delete the temp GZip file
                    if (totalFileSize < uploadFileSize) {
                        uploadFileSize = totalFileSize;
                        uploadFilePath = parameters.file;
                        isGzip = false;
                        tmpFile.cleanup();
                    }
                    let abortFileUpload = false;
                    // upload only a single chunk at a time
                    while (offset < uploadFileSize) {
                        const chunkSize = Math.min(uploadFileSize - offset, parameters.maxChunkSize);
                        if (abortFileUpload) {
                            // if we don't want to continue in the event of an error, any pending upload chunks will be marked as failed
                            failedChunkSizes += chunkSize;
                            continue;
                        }
                        // if an individual file is greater than 100MB (1024*1024*100) in size, display extra information about the upload status
                        if (uploadFileSize > 104857600) {
                            this.statusReporter.updateLargeFileStatus(parameters.file, offset, uploadFileSize);
                        }
                        const start = offset;
                        const end = offset + chunkSize - 1;
                        offset += parameters.maxChunkSize;
                        const result = yield this.uploadChunk(httpClientIndex, parameters.resourceUrl, fs.createReadStream(uploadFilePath, {
                            start,
                            end,
                            autoClose: false
                        }), start, end, uploadFileSize, isGzip, totalFileSize);
                        if (!result) {
                            // Chunk failed to upload, report as failed and do not continue uploading any more chunks for the file. It is possible that part of a chunk was
                            // successfully uploaded so the server may report a different size for what was uploaded
                            isUploadSuccessful = false;
                            failedChunkSizes += chunkSize;
                            core_1.warning(`Aborting upload for ${parameters.file} due to failure`);
                            abortFileUpload = true;
                        }
                    }
                }))
                    .then(() => __awaiter(this, void 0, void 0, function* () {
                    // only after the file upload is complete and the temporary file is deleted, return the UploadResult
                    return new Promise(resolve => {
                        resolve({
                            isSuccess: isUploadSuccessful,
                            successfullUploadSize: uploadFileSize - failedChunkSizes,
                            totalSize: totalFileSize
                        });
                    });
                }));
            }
        });
    }
    /**
     * Uploads a chunk of an individual file to the specified resourceUrl. If the upload fails and the status code
     * indicates a retryable status, we try to upload the chunk as well
     * @param {number} httpClientIndex The index of the httpClient being used to make all the necessary calls
     * @param {string} resourceUrl Url of the resource that the chunk will be uploaded to
     * @param {NodeJS.ReadableStream} data Stream of the file that will be uploaded
     * @param {number} start Starting byte index of file that the chunk belongs to
     * @param {number} end Ending byte index of file that the chunk belongs to
     * @param {number} uploadFileSize Total size of the file in bytes that is being uploaded
     * @param {boolean} isGzip Denotes if we are uploading a Gzip compressed stream
     * @param {number} totalFileSize Original total size of the file that is being uploaded
     * @returns if the chunk was successfully uploaded
     */
    uploadChunk(httpClientIndex, resourceUrl, data, start, end, uploadFileSize, isGzip, totalFileSize) {
        return __awaiter(this, void 0, void 0, function* () {
            // prepare all the necessary headers before making any http call
            const requestOptions = utils_1.getRequestOptions('application/octet-stream', true, isGzip, totalFileSize, end - start + 1, utils_1.getContentRange(start, end, uploadFileSize));
            const uploadChunkRequest = () => __awaiter(this, void 0, void 0, function* () {
                const client = this.uploadHttpManager.getClient(httpClientIndex);
                return yield client.sendStream('PUT', resourceUrl, data, requestOptions);
            });
            let retryCount = 0;
            const retryLimit = config_variables_1.getUploadRetryCount();
            // allow for failed chunks to be retried multiple times
            while (retryCount <= retryLimit) {
                try {
                    const response = yield uploadChunkRequest();
                    // Always read the body of the response. There is potential for a resource leak if the body is not read which will
                    // result in the connection remaining open along with unintended consequences when trying to dispose of the client
                    yield response.readBody();
                    if (utils_1.isSuccessStatusCode(response.message.statusCode)) {
                        return true;
                    }
                    else if (utils_1.isRetryableStatusCode(response.message.statusCode)) {
                        retryCount++;
                        if (retryCount > retryLimit) {
                            core_1.info(`Retry limit has been reached for chunk at offset ${start} to ${resourceUrl}`);
                            return false;
                        }
                        else {
                            core_1.info(`HTTP ${response.message.statusCode} during chunk upload, will retry at offset ${start} after ${config_variables_1.getRetryWaitTimeInMilliseconds} milliseconds. Retry count #${retryCount}. URL ${resourceUrl}`);
                            this.uploadHttpManager.disposeAndReplaceClient(httpClientIndex);
                            yield new Promise(resolve => setTimeout(resolve, config_variables_1.getRetryWaitTimeInMilliseconds()));
                        }
                    }
                    else {
                        core_1.info(`#ERROR# Unable to upload chunk to ${resourceUrl}`);
                        // eslint-disable-next-line no-console
                        console.log(response);
                        return false;
                    }
                }
                catch (error) {
                    // eslint-disable-next-line no-console
                    console.log(error);
                    retryCount++;
                    if (retryCount > retryLimit) {
                        core_1.info(`Retry limit has been reached for chunk at offset ${start} to ${resourceUrl}`);
                        return false;
                    }
                    else {
                        core_1.info(`Retrying chunk upload after encountering an error`);
                        this.uploadHttpManager.disposeAndReplaceClient(httpClientIndex);
                        yield new Promise(resolve => setTimeout(resolve, config_variables_1.getRetryWaitTimeInMilliseconds()));
                    }
                }
            }
            return false;
        });
    }
    /**
     * Updates the size of the artifact from -1 which was initially set when the container was first created for the artifact.
     * Updating the size indicates that we are done uploading all the contents of the artifact
     */
    patchArtifactSize(size, artifactName) {
        return __awaiter(this, void 0, void 0, function* () {
            const requestOptions = utils_1.getRequestOptions('application/json', false, false);
            const resourceUrl = new url_1.URL(utils_1.getArtifactUrl());
            resourceUrl.searchParams.append('artifactName', artifactName);
            const parameters = { Size: size };
            const data = JSON.stringify(parameters, null, 2);
            core_1.debug(`URL is ${resourceUrl.toString()}`);
            // use the first client from the httpManager, `keep-alive` is not used so the connection will close immediatly
            const client = this.uploadHttpManager.getClient(0);
            const rawResponse = yield client.patch(resourceUrl.toString(), data, requestOptions);
            const body = yield rawResponse.readBody();
            if (utils_1.isSuccessStatusCode(rawResponse.message.statusCode)) {
                core_1.debug(`Artifact ${artifactName} has been successfully uploaded, total size ${size}`);
            }
            else if (rawResponse.message.statusCode === 404) {
                throw new Error(`An Artifact with the name ${artifactName} was not found`);
            }
            else {
                // eslint-disable-next-line no-console
                console.log(body);
                throw new Error(`Unable to finish uploading artifact ${artifactName}`);
            }
        });
    }
}
exports.UploadHttpClient = UploadHttpClient;
//# sourceMappingURL=upload-http-client.js.map